/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include "cone.h"

#include <glint/glint.h>

#include <iostream>

using namespace glint;
using namespace std;

void
usage()
{
    cerr << "usage: uniform [ vertex_shader_file fragment_shader_file ]" << endl;
    exit(2);
}

int
main(int argc, const char* argv[])
{
    string vertex_file("vertex.glsl");
    string fragment_file("fragment.glsl");
    if (argc != 1 && argc != 3)
    {
        usage();
    }
    if (argc == 3)
    {
        vertex_file = argv[1];
        fragment_file = argv[2];
    }

    try
    {
        auto rt = Runtime::create();
        rt->create_display(300, 200);

        glDisable(GL_BLEND);
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LEQUAL);
        glEnable(GL_CULL_FACE);
        glFrontFace(GL_CCW);
        glCullFace(GL_BACK);

        Program::UPtr prog;
        {
            auto vs = VertexShader::create_from_file(vertex_file);
            auto fs = FragmentShader::create_from_file(fragment_file);
            prog = Program::create(*vs, *fs);
        }
        prog->validate();
        prog->use();

        prog->uniform("lightColor[0]") = FloatVec3({ { 1, 0, 0} });
        prog->uniform("lightColor[1]") = FloatVec3({ { 0, 1, 0} });
        prog->uniform("lightColor[2]") = FloatVec3({ { 0, 0, 1} });

        cout << prog->uniform("lightPosition").name() << endl;
        cout << prog->uniform("lightPosition").type() << endl;
        cout << prog->uniform("lightPosition").size() << endl;

        constexpr FloatVec3 cameraPos = { { 0, 0, -4 } };
        prog->uniform("cameraPosition") = cameraPos;

        constexpr FloatMat4 mvp = { {
                                       { { 1, 0, 0, 0 } },
                                       { { 0, 1, 0, 0 } },
                                       { { 0, 0, 1, 0 } },
                                       { { cameraPos[0], cameraPos[1], cameraPos[2], 1 } }
                                  } };
        prog->uniform("mvp") = mvp;


        // TODO: why does this throw?
        //FloatVec3Array& lpos = dynamic_cast<FloatVec3Array&>(prog->uniform("lightPosition"));

        UniformVariable& lightPosition = prog->uniform("lightPosition");

        constexpr int num_lights = 3;
        FloatVec3Array lpos;
        lpos.resize(num_lights);
        for (int i = 0; i < num_lights; ++i)
        {
            constexpr float radius = 1.75f;
            float r = ((2 * M_PI) / num_lights) * i;
            lpos[i][0] = cos(r) * radius;
            lpos[i][1] = cos(r) * sin(r);
            lpos[i][2] = sin(r) * radius;
        }
        lightPosition = lpos;

        FloatVec3Array vertices;
        FloatVec3Array normals;
        int cone_segments = 3;
        create_cone(cone_segments, vertices, normals);

        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(vertices[0]), &vertices[0]);
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(normals[0]), &normals[0]);
        cout << sizeof(vertices[0]) << endl;
        check_gl("glVertexAttribPointer() failed");

        glEnableVertexAttribArray(0);
        glEnableVertexAttribArray(1);
        check_gl("glEnableVertexAttribArray() failed");

        glBindAttribLocation(prog->id(), 0, "vertex");
        glBindAttribLocation(prog->id(), 1, "normal");
        check_gl("glBindAttribLocation() failed");

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glDrawArrays(GL_TRIANGLE_FAN, 0, vertices.size());
        check_gl("glDrawArrays() failed");

        rt->display()->swap_buffers();
        sleep(3);
    }
    catch (const Exception& e)
    {
        cerr << e.to_string() << endl;
        exit(1);
    }
}
