/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_ATTRIBUTEVARIABLE_H
#define GLINT_ATTRIBUTEVARIABLE_H

#include <glint/ShaderVariable.h>
#include <glint/Types.h>

namespace glint
{

namespace internal
{

class AttributeVariableImpl;

} // namespace internal

//
// AttributeVariable is the common ancestor of all attributes. This allows us to polymorphically
// store attributes, for example, as the look-up value in a map.
//
// AttributeVariable provides assignment functionality because shader attributes are used
// mostly as write-only variables. This saves the caller having to first cast to
// the derived type with a dynamic_cast. As long as the value being assigned matches the
// type of the attribute at run time, the assignment will work. Otherwise, the caller
// gets a std::bad_cast exception.
//
// To read the value of an attribute (used rarely), the caller must down-cast to the concrete type.
//

class AttributeVariable : public ShaderVariable
{
public:
    enum class ArrayState : GLint { disabled, enabled };

    ArrayState array_state() const;
    void set_array_state(ArrayState state);

    void set(GLfloat rhs);
    AttributeVariable& operator=(GLfloat rhs);

    void set(const FloatVec2& rhs);
    AttributeVariable& operator=(const FloatVec2& rhs);

    void set(const FloatVec3& rhs);
    AttributeVariable& operator=(const FloatVec3& rhs);

    void set(const FloatVec4& rhs);
    AttributeVariable& operator=(const FloatVec4& rhs);

    void set(const FloatMat2& rhs);
    AttributeVariable& operator=(const FloatMat2& rhs);

    void set(const FloatMat3& rhs);
    AttributeVariable& operator=(const FloatMat3& rhs);

    void set(const FloatMat4& rhs);
    AttributeVariable& operator=(const FloatMat4& rhs);

    virtual ~AttributeVariable();

protected:
    AttributeVariable(const std::shared_ptr<internal::AttributeVariableImpl>& derived);
};

//
// One derived class for each possible type of variable.
//

class FloatAttribute : public AttributeVariable
{
public:
    FloatAttribute(GLuint prog_id, GLint location, const std::string& name);
    virtual ~FloatAttribute();

    GLfloat get() const;
    void set(GLfloat v);

    operator GLfloat() const;
    FloatAttribute& operator=(GLfloat rhs);
    FloatAttribute& operator=(const FloatAttribute& rhs);
};

class FloatVec2Attribute : public AttributeVariable
{
public:
    FloatVec2Attribute(GLuint prog_id, GLint location, const std::string& name);
    virtual ~FloatVec2Attribute();

    FloatVec2 get() const;
    void get(GLfloat& x, GLfloat& y) const;
    void set(const FloatVec2& v);
    void set(GLfloat x, GLfloat y);

    operator FloatVec2() const;
    FloatVec2Attribute& operator=(const FloatVec2& rhs);
    FloatVec2Attribute& operator=(const FloatVec2Attribute& rhs);
};

class FloatVec3Attribute : public AttributeVariable
{
public:
    FloatVec3Attribute(GLuint prog_id, GLint location, const std::string& name);
    virtual ~FloatVec3Attribute();

    FloatVec3 get() const;
    void get(GLfloat& x, GLfloat& y, GLfloat& z) const;
    void set(const FloatVec3& v);
    void set(GLfloat x, GLfloat y, GLfloat z);

    operator FloatVec3() const;
    FloatVec3Attribute& operator=(const FloatVec3& rhs);
    FloatVec3Attribute& operator=(const FloatVec3Attribute& rhs);
};

class FloatVec4Attribute : public AttributeVariable
{
public:
    FloatVec4Attribute(GLuint prog_id, GLint location, const std::string& name);
    virtual ~FloatVec4Attribute();

    FloatVec4 get() const;
    void get(GLfloat& x, GLfloat& y, GLfloat& z, GLfloat& w) const;
    void set(const FloatVec4& v);
    void set(GLfloat x, GLfloat y, GLfloat z, GLfloat w);

    operator FloatVec4() const;
    FloatVec4Attribute& operator=(const FloatVec4& rhs);
    FloatVec4Attribute& operator=(const FloatVec4Attribute& rhs);
};

class FloatMat2Attribute : public AttributeVariable
{
public:
    FloatMat2Attribute(GLuint prog_id, GLint location, const std::string& name);
    virtual ~FloatMat2Attribute();

    FloatMat2 get() const;
    void set(const FloatMat2& v);

    operator FloatMat2() const;
    FloatMat2Attribute& operator=(const FloatMat2& rhs);
    FloatMat2Attribute& operator=(const FloatMat2Attribute& rhs);
};

class FloatMat3Attribute : public AttributeVariable
{
public:
    FloatMat3Attribute(GLuint prog_id, GLint location, const std::string& name);
    virtual ~FloatMat3Attribute();

    FloatMat3 get() const;
    void set(const FloatMat3& v);

    operator FloatMat3() const;
    FloatMat3Attribute& operator=(const FloatMat3& rhs);
    FloatMat3Attribute& operator=(const FloatMat3Attribute& rhs);
};

class FloatMat4Attribute : public AttributeVariable
{
public:
    FloatMat4Attribute(GLuint prog_id, GLint location, const std::string& name);
    virtual ~FloatMat4Attribute();

    FloatMat4 get() const;
    void set(const FloatMat4& v);

    operator FloatMat4() const;
    FloatMat4Attribute& operator=(const FloatMat4& rhs);
    FloatMat4Attribute& operator=(const FloatMat4Attribute& rhs);
};

} // namespace glint

#endif
