/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/Program.h>
#include <glint/Shader.h>
#include <glint/Runtime.h>
#include <glint/GlintExceptions.h>

#include <gtest/gtest.h>
#include <boost/regex.hpp>  // BUGFIX: We don't use std::regex because, as of gcc 4.7.2, it is still broken

using namespace std;
using namespace glint;

TEST(Attribute, attribute_lookup)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "attribute float my_float;"
                           "void main()"
                           "{"
                           "    gl_Position = vec4(my_float);"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    auto v = VertexShader::create(v_source);
    auto f = FragmentShader::create(f_source);
    auto p = Program::create(*v, *f);

    EXPECT_EQ(1, p->attributes().size());

    {
        AttributeVariable& uv = p->attribute("my_float");
        FloatAttribute& fa = dynamic_cast<FloatAttribute&>(uv);
        EXPECT_NE(-1, fa.location());
        EXPECT_EQ("my_float", fa.name());
        EXPECT_EQ(GL_FLOAT, fa.type());
        EXPECT_EQ(1, fa.size());
        EXPECT_EQ(AttributeVariable::ArrayState::disabled, fa.array_state());
    }

    // Getting and setting an attribute before the program is installed is OK.
    AttributeVariable& uv = p->attribute("my_float");
    FloatAttribute& iu = dynamic_cast<FloatAttribute&>(uv);
    iu.set(1.0f);
    iu.get();
    p->use();

    try
    {
        p->uniform("Michi");
        FAIL();
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: no uniform variable \"Michi\" found for program id [0-9]+",
                       boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }

    uv.set_array_state(AttributeVariable::ArrayState::disabled);
    EXPECT_EQ(AttributeVariable::ArrayState::disabled, uv.array_state());
    uv.set_array_state(AttributeVariable::ArrayState::enabled);
    EXPECT_EQ(AttributeVariable::ArrayState::enabled, uv.array_state());
    uv.set_array_state(AttributeVariable::ArrayState::disabled);
    EXPECT_EQ(AttributeVariable::ArrayState::disabled, uv.array_state());
}

TEST(Attribute, attributes)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    // Shader with all possible types of attribute. Silly use of these in the body of the shader, so they
    // are not optimized out of existence.
    const char* v_source = "attribute float f;"
                           "attribute vec2 fv2;"
                           "attribute vec3 fv3;"
                           "attribute vec4 fv4;"
                           "attribute mat2 fm2;"
                           "attribute mat3 fm3;"
                           "attribute mat4 fm4;"
                           "void main()"
                           "{"
                           "    gl_Position = vec4(f + fv2[1] + fv3[2] + fv4[3]);"
                           "    gl_Position = gl_Position + vec4(fm2[1][1] + fm3[2][2] + fm4[3][3]);"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    auto v = VertexShader::create(v_source);
    auto f = FragmentShader::create(f_source);
    auto p = Program::create(*v, *f);

    {
        AttributeVariable& av = p->attribute("f");
        av.set(2.0f);
        FloatAttribute& a = dynamic_cast<FloatAttribute&>(av);
        EXPECT_EQ(2.0f, a.get());
        a.set(3.0f);
        EXPECT_EQ(3.0f, a.get());
    }

    {
        FloatVec2 in = { { 1.0f, 2.0f } };
        AttributeVariable& av = p->attribute("fv2");
        av.set(in);
        FloatVec2Attribute& a = dynamic_cast<FloatVec2Attribute&>(av);
        FloatVec2 r = a.get();
        in[1] = 3.0f;
        a.set(in);
        r = a.get();
        EXPECT_EQ(in, r);
    }

    {
        FloatVec3 in = { { 1.0f, 2.0f, 3.0f } };
        AttributeVariable& av = p->attribute("fv3");
        av.set(in);
        FloatVec3Attribute& a = dynamic_cast<FloatVec3Attribute&>(av);
        FloatVec3 r = a.get();
        EXPECT_EQ(in, r);
        in[2] = 4.0f;
        a.set(in);
        r = a.get();
        EXPECT_EQ(in, r);
    }

    {
        FloatVec4 in = { { 1.0f, 2.0f, 3.0f, 4.0f } };
        AttributeVariable& av = p->attribute("fv4");
        av.set(in);
        FloatVec4Attribute& a = dynamic_cast<FloatVec4Attribute&>(av);
        FloatVec4 r = a.get();
        EXPECT_EQ(in, r);
        in[3] = 5.0f;
        a.set(in);
        r = a.get();
        EXPECT_EQ(in, r);
    }

    {
        FloatMat2 in = { {
                          { { 1.0f, 2.0f } },
                          { { 3.0f, 4.0f } }
                       } };
        AttributeVariable& av = p->attribute("fm2");
        av.set(in);
        FloatMat2Attribute& a = dynamic_cast<FloatMat2Attribute&>(av);
        FloatMat2 r = a.get();
        EXPECT_EQ(in, r);
        in[1][1] = 5.0f;
        a.set(in);
        r = a.get();
        EXPECT_EQ(in, r);
    }

    {
        FloatMat3 in = { {
                          { { 1.0f, 2.0f, 3.0f } },
                          { { 4.0f, 5.0f, 6.0f } },
                          { { 7.0f, 8.0f, 9.0f } }
                       } };
        AttributeVariable& av = p->attribute("fm3");
        av.set(in);
        FloatMat3Attribute& a = dynamic_cast<FloatMat3Attribute&>(av);
        FloatMat3 r = a.get();
        EXPECT_EQ(in, r);
        in[2][2] = 10.0f;
        a.set(in);
        r = a.get();
        EXPECT_EQ(in, r);
    }

    {
        FloatMat4 in = { {
                          { {  1.0f,  2.0f,  3.0f,  4.0f } },
                          { {  5.0f,  6.0f,  7.0f,  8.0f } },
                          { {  9.0f, 10.0f, 11.0f, 12.0f } },
                          { { 13.0f, 14.0f, 15.0f, 16.0f } },
                       } };
        AttributeVariable& av = p->attribute("fm4");
        av.set(in);
        FloatMat4Attribute& a = dynamic_cast<FloatMat4Attribute&>(av);
        FloatMat4 r = a.get();
        EXPECT_EQ(in, r);
        in[3][3] = 17.0f;
        a.set(in);
        r = a.get();
        EXPECT_EQ(in, r);
    }
}
