/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/Program.h>
#include <glint/Shader.h>
#include <glint/Runtime.h>
#include <glint/GlintExceptions.h>

#include <gtest/gtest.h>

using namespace std;
using namespace glint;

TEST(Program, basic)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "void main()"
                           "{"
                           "    gl_Position = vec4(0);"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    auto v = VertexShader::create(v_source);
    auto f = FragmentShader::create(f_source);
    auto p = Program::create(*v, *f);

    EXPECT_NE(0, p->id());
    EXPECT_EQ(v->id(), p->vertex_shader_id());
    EXPECT_EQ(f->id(), p->fragment_shader_id());

    EXPECT_FALSE(p->in_use());
    p->use();
    EXPECT_TRUE(p->in_use());

    EXPECT_TRUE(p->uniforms().empty());
}

TEST(Program, bad_vertex_shader)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "void main()"
                           "{"
                           "/* Invalid, no output */"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    // We deliberately enclose the entire sequence in the try block because
    // at what stage (compile, link, or use) things fail depends on the implementation.
    try
    {
        auto v = VertexShader::create(v_source);
        auto f = FragmentShader::create(f_source);
        auto p = Program::create(*v, *f);
        p->use();
        FAIL();
    }
    catch (const OpenGLException& e)
    {
        // No string comparison here for the error message because the message is implementation-dependent
    }
}

TEST(Program, bad_link)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "float foo() { return 1.0; }";

    const char* f_source = "float foo() { return 2.0; }";

    // We deliberately enclose the entire sequence in the try block because
    // at what stage (compile, link, or use) things fail depends on the implementation.
    try
    {
        auto v = VertexShader::create(v_source);
        auto f = FragmentShader::create(f_source);
        auto p = Program::create(*v, *f);
        p->use();
        FAIL();
    }
    catch (const Exception& e)
    {
        // No string comparison here for the error message because the message is implementation-dependent
    }
}

TEST(Program, assignment) // TODO: finish this
{
}

#if 0
TEST(Program, use_change)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "uniform float my_float;"
                           "void main()"
                           "{"
                           "    gl_Position = vec4(my_float);"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    auto v = VertexShader::create(v_source);
    auto f = FragmentShader::create(f_source);
    auto p = Program::create(*v, *f);

    // Setting a variable before the program is installed should fail, but reading it is OK.
    try
    {
        UniformVariable& uv = p->uniform("my_float");
        FloatArrayUniform& iu = dynamic_cast<FloatArrayUniform&>(uv);
        iu.set({ { 1.0f, 2.0f }});
        FAIL();
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: cannot set uniform variable \"my_float\" at location [0-9]+;"
                       " size = 2, program id = [0-9]+: program is not currently in use",
                       boost::regex::egrep);
        cout << e.to_string() << endl;
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }

    {
        UniformVariable& uv = p->uniform("my_float");
        FloatArrayUniform& iu = dynamic_cast<FloatArrayUniform&>(uv);
        iu.get();
    }

    p->use();

    {
        UniformVariable& uv = p->uniform("my_float");
        FloatArrayUniform& iu = dynamic_cast<FloatArrayUniform&>(uv);
        EXPECT_EQ(p->id(), iu.program_id());
        EXPECT_NE(-1, iu.location());
        EXPECT_EQ("my_float", iu.name());
        EXPECT_EQ(GL_FLOAT, iu.type());
        EXPECT_EQ(2, iu.size());
    }

    // Repeat for the same variable, but this time with an array name.
    {
        UniformVariable& uv = p->uniform("my_float[0]");
        FloatUniform& iu = dynamic_cast<FloatUniform&>(uv);
        EXPECT_EQ(p->id(), iu.program_id());
        EXPECT_NE(-1, iu.location());
        EXPECT_EQ("my_float[0]", iu.name());
        EXPECT_EQ(GL_FLOAT, iu.type());
        EXPECT_EQ(1, iu.size());
    }

    // Once more for coverage testing. This time, the variable should come out of the cache.
    {
        UniformVariable& uv = p->uniform("my_float[0]");
        FloatUniform& iu = dynamic_cast<FloatUniform&>(uv);
        EXPECT_EQ(p->id(), iu.program_id());
        EXPECT_NE(-1, iu.location());
        EXPECT_EQ("my_float[0]", iu.name());
        EXPECT_EQ(GL_FLOAT, iu.type());
        EXPECT_EQ(1, iu.size());
    }

    try
    {
        p->uniform("Michi");
        FAIL();
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: no uniform variable \"Michi\" found for program id [0-9]+",
                       boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }
}
#endif
